/*
 * deviced
 *
 * Copyright (c) 2012 - 2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <fcntl.h>
#include <signal.h>
#include <dirent.h>
#include <ctype.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <poll.h>
#include <mntent.h>
#include <system_info.h>
#include <sys/mount.h>
#include "log.h"
#include "common.h"

#define PERMANENT_DIR   "/tmp/permanent"
#define VIP_DIR         "/tmp/vip"
#define BUFF_MAX        255

#define APP_ATTR_PATH "/proc/%d/attr/current"

/**
 * Opens "/proc/$pid/oom_score_adj" file for w/r;
 * Return: FILE pointer or NULL
 */
FILE *open_proc_oom_score_adj_file(int pid, const char *mode)
{
	char buf[32];
	FILE *fp;

	snprintf(buf, sizeof(buf), "/proc/%d/oom_score_adj", pid);
	fp = fopen(buf, mode);
	return fp;
}

int get_cmdline_name(pid_t pid, char *cmdline, size_t cmdline_size)
{
	int fd, ret;
	char buf[PATH_MAX + 1];
	char *filename;

	snprintf(buf, sizeof(buf), "/proc/%d/cmdline", pid);
	fd = open(buf, O_RDONLY);
	if (fd < 0) {
		errno = ESRCH;
		return -1;
	}

	ret = read(fd, buf, PATH_MAX);
	close(fd);
	if (ret < 0)
		return -1;

	buf[PATH_MAX] = '\0';

	filename = strrchr(buf, '/');
	if (filename == NULL)
		filename = buf;
	else
		filename = filename + 1;

	if (cmdline_size < strlen(filename) + 1) {
		errno = EOVERFLOW;
		return -1;
	}

	strncpy(cmdline, filename, cmdline_size - 1);
	cmdline[cmdline_size - 1] = '\0';
	return 0;
}

int is_vip(int pid)
{
	if (pid < 1)
		return -1;

	char buf[PATH_MAX];

	snprintf(buf, PATH_MAX, "%s/%d", VIP_DIR, pid);

	if (access(buf, R_OK) == 0)
		return 1;
	else
		return 0;
}

/*
 * Helper function
 * - Read from sysfs entry
 * - Write to sysfs entry
 */
int sys_check_node(char *path)
{
	int fd;

	fd = open(path, O_RDONLY);
	if (fd == -1)
		return -1;

	close(fd);
	return 0;
}

static int sys_read_buf(char *file, char *buf)
{
	int fd;
	int r;
	int ret = 0;

	fd = open(file, O_RDONLY);
	if (fd == -1)
		return -ENOENT;

	r = read(fd, buf, BUFF_MAX);
	if ((r >= 0) && (r < BUFF_MAX))
		buf[r] = '\0';
	else
		ret = -EIO;

	close(fd);

	return ret;
}

static int sys_write_buf(char *file, char *buf)
{
	int fd;
	int r;
	int ret = 0;

	fd = open(file, O_WRONLY);
	if (fd == -1)
		return -ENOENT;

	r = write(fd, buf, strlen(buf));
	if (r < 0)
		ret = -EIO;

	close(fd);

	return ret;
}

int sys_get_int(char *fname, int *val)
{
	char buf[BUFF_MAX];
	int ret = 0;

	if (sys_read_buf(fname, buf) == 0) {
		*val = atoi(buf);
	} else {
		*val = -1;
		ret = -EIO;
	}

	return ret;
}

int sys_set_int(char *fname, int val)
{
	char buf[BUFF_MAX];
	int ret = 0;

	snprintf(buf, sizeof(buf), "%d", val);

	if (sys_write_buf(fname, buf) != 0)
		ret = -EIO;

	return ret;
}

int sys_get_str(char *fname, char *str)
{
	char buf[BUFF_MAX] = {0};

	if (sys_read_buf(fname, buf) == 0) {
		strncpy(str, buf, strlen(buf)+1);
		return 0;
	}

	return -1;
}

int sys_set_str(char *fname, char *val)
{
	int r = -1;

	if (val != NULL) {
		if (sys_write_buf(fname, val) == 0)
			r = 0;
	}

	return r;
}

static int terminate_processes_on_partition(const char *partition, bool force)
{
	const char *argv[7] = {"/usr/bin/fuser", "-m", "-k", "-s", NULL, NULL, NULL};
	int argc;

	if (force)
		argv[4] = "-SIGKILL";
	else
		argv[4] = "-SIGTERM";

	argv[5] = partition;
	argc = sizeof(argv) / sizeof(argv[0]);

	return run_child(argc, argv);
}

int mount_check(const char *path)
{
	int ret = false;
	struct mntent *mnt;
	const char *table = "/etc/mtab";
	FILE *fp;
	int len;

	fp = setmntent(table, "r");
	if (!fp)
		return ret;

	len = strlen(path) + 1;
	while (1) {
		mnt = getmntent(fp);
		if (mnt == NULL)
			break;
		if (!strncmp(mnt->mnt_dir, path, len)) {
			ret = true;
			break;
		}
	}
	endmntent(fp);
	return ret;
}

static int get_num_processes_on_partition(const char *part_path)
{
	FILE *fp;
	char *cmd = NULL;
	char *line = NULL;
	size_t len = 0;
	int num_processes = 0;
	int ret = 0;

	ret = asprintf(&cmd, "fuser -m %s | grep -o '[0-9]*'", part_path);
	if (ret < 0)
		return -1;

	printf("cmd=%s\n", cmd);
	fp = popen(cmd, "r");
	free(cmd);
	if (fp == NULL)
		return -1;

	while (getline(&line, &len, fp) != -1) {
		num_processes++;
	}

	free(line);
	pclose(fp);

	return num_processes;
}

void umount_partition_by_kill(const char *path, const int max_retry)
{
	FILE *fp;
	char *part_path = NULL;
	size_t len = 0;
	int ret = 0, retry = 0;
	int remain;
	char *cmd = NULL;

	/* if the path is not a mountpoint, do not perform umounting and killing */
	if (!mount_check(path))
		return;

	ret = asprintf(&cmd, "mount | grep \" on %s \" | awk '{print $1}'", path);
	if (ret < 0)
		return;

	fp = popen(cmd, "r");
	free(cmd);
	if (!fp)
		return;

	ret = getline(&part_path, &len, fp);
	if (ret == -1 || !part_path) {
		pclose(fp);
		return;
	} else if (ret > 0 && *(part_path + ret - 1) == '\n') {
		*(part_path + ret -1) = '\0';
	}

	umount2(path, MNT_DETACH);

	do {
		sync();

		/* Kill processes with SIGTERM */
		terminate_processes_on_partition(part_path, false);
		usleep((useconds_t)MSEC_TO_USEC(500));

		/* Kill processes with SIGKILL */
		terminate_processes_on_partition(part_path, true);
		usleep((useconds_t)MSEC_TO_USEC(200));

		remain = get_num_processes_on_partition(part_path);

		retry++;

	} while (remain > 0 && retry < max_retry);

	pclose(fp);
	return;
}

void print_time(const char *prefix)
{
	struct timeval tv;
	struct tm tm;
	struct tm *ret;
	gettimeofday(&tv, NULL);
	ret = localtime_r(&(tv.tv_sec), &tm);
	if (ret)
		_D("%s --> %d:%02d:%02d %d",
				prefix, tm.tm_hour, tm.tm_min, tm.tm_sec, tv.tv_usec);
}

int get_privilege(pid_t pid, char *name, size_t len)
{
	char path[PATH_MAX];
	char attr[BUFF_MAX];
	size_t attr_len;
	FILE *fp;

	snprintf(path, sizeof(path), APP_ATTR_PATH, pid);

	fp = fopen(path, "r");
	if (!fp)
		return -errno;

	attr_len = fread(attr, 1, sizeof(attr) - 1, fp);
	fclose(fp);
	if (attr_len == 0)
		return -ENOENT;

	attr[attr_len] = '\0';

	snprintf(name, len, "%s", attr);
	return 0;
}

#define MODEL_NAME      "http://tizen.org/system/model_name"
#define MODEL_EMULATOR  "Emulator"

bool is_emulator(void)
{
	int ret;
	char *model_name = NULL;
	static bool emul = false;
	static int set = 0;

	if (set)
		return emul;

	ret = system_info_get_platform_string(MODEL_NAME, &model_name);
	if (ret < 0) {
		_E("Cannot get model name(%d)", ret);
		return emul;
	}

	if (!strncmp(MODEL_EMULATOR, model_name, strlen(model_name) + 1))
		emul = true;

	set = 1;
	free(model_name);

	return emul;
}
